/*********************************************************
    Module Name: FrameMgr.c
    Module Date: 04/10/2004
    Module Auth: John Orlando 
    Copyright (c) 2004 John Orlando  All Rights Reserved 

    Description: This modules is responsible for performing
    both medium and high level processing on image data.
    This is performed at both the line level as well as
    the frame level.  It controls the main flow of the
    system, adhering to all the critical timing
    requirements (such as when serial data can be transferred,
    etc).
    *********************************************************/

/*	Includes */
#include <stdlib.h>
#include <string.h>
#include <avr/io.h>
#include "Executive.h"
#include "UIMgr.h"
#include "FrameMgr.h"
#include "CamInterface.h"
#include "UartInterface.h"
#include "Utility.h"
#include "I2CInterface.h"
#include "CamConfig.h"
#include "CommonDefs.h"

/* 	Local Structures and Typedefs */
enum
{
    ST_FrameMgr_idle,
    ST_FrameMgr_TrackingFrame,
    ST_FrameMgr_DumpingFrame
};

typedef unsigned char FrameMgr_State_t;

/*  Definitions */
/* The most objects that can be tracked at any one time is 8.
    This number is determined by the number of bytes that can be
    sent out during a frame (one byte per line, 144 lines per frame)
    with the number of bytes in a tracked object (7) + some wiggle
    room :-) ... I guess this could be increased to around 20 if
    we had enough room and cycles to process objects between lines */
#define MAX_TRACKED_OBJECTS	8

/* This defines the number of bytes that make up a trackedObject_t
    structure...we don't just use sizeof() because there is a fill
    byte in there to cause the structure to align properly */
#define SIZE_OF_TRACKED_OBJECT 7

/* This define is used to turn off the timer overflow interrupt
    that is generated when the PCLK overflows TIMER1 */
#define DISABLE_PCLK_TIMER1_OVERFLOW_BITMASK 0xFB

/* This enum describes the possible colors that can
    be tracked by the system.  This can't be represented as
    simple color names (red, brown, etc) due to the fact that
    the user sets which colors will be associated with which
    bits.  Remember...after the AND operation of the indexed
    color map values executes, either a single bit indicating
    the color should be set, or no bits indicating that the
    color isn't represented in the color map (notTracked). */
enum
{
    notTracked,
    color1,		/* bit 1 color */
    color2,             /* bit 2 color */
    color3,             /* bit 3 color */
    color4,             /* bit 4 color */
    color5,             /* bit 5 color */
    color6,             /* bit 6 color */
    color7,             /* bit 7 color */
    color8              /* bit 8 color */
};

typedef unsigned char trackedColor_t;

/* This structure defines the info that needs to be
    maintained for each trackedObject in the trackingTable */
typedef struct
{
    trackedColor_t  color;
    unsigned char lastLineXStart;
    unsigned char lastLineXFinish;
    unsigned char x_upperLeft;
    unsigned char y_upperLeft;
    unsigned char x_lowerRight;
    unsigned char y_lowerRight;
    unsigned char fillByte;
} trackedObject_t;

/*  Local Variables */
/* The trackedObjectTable is used to hold up to eight tracked objects
    while they are being acquired. */
static trackedObject_t trackedObjectTable[MAX_TRACKED_OBJECTS];
static trackedObject_t *pCurrentTrackedObjectTable = trackedObjectTable;
static unsigned char lineCount = 0;
static FrameMgr_State_t currentState = ST_FrameMgr_idle;
static unsigned char numCurrTrackedObjects = 0;
static unsigned char numPrevTrackedObjects = 0;
static unsigned char trackedLineCount = 0;

/*  Local Functions  */
static void FrameMgr_findConnectedness(void);

/*  Extern Functions */
/* These functions are located in assembly files, and thus
    must be externed here so they can be referenced in the source below. */
extern void CamIntAsm_waitForNewTrackingFrame(unsigned char *pBuffer, unsigned char *pMemLookup);
extern void CamIntAsm_waitForNewDumpFrame(unsigned char *pCurrBuffer, unsigned char *pPrevBuffer);
extern void CamIntAsm_acquireTrackingLine(unsigned char *pBuffer, unsigned char *pMemLookup);
extern void CamIntAsm_acquireDumpLine(unsigned char *pCurrBuffer, unsigned char *pPrevBuffer);

/***********************************************************
    Function Name: FrameMgr_init
    Function Description: This function is responsible
    for initializing the FrameMgr.  This includes
    setting up the various buffers and data needed to
    process each frame of image data.
    Inputs:  none
    Outputs: none
    ***********************************************************/
void FrameMgr_init(void)
{
    memset(trackedObjectTable,0x00,sizeof(trackedObjectTable));
}


/***********************************************************
    Function Name: FrameMgr_dispatchEvent
    Function Description: This function is responsible for
    taking an incoming event and performing the needed
    actions with it as pertains to the FrameMgr.
    Inputs:  event - the generated event
    Outputs: none
    ***********************************************************/
void FrameMgr_dispatchEvent(unsigned char event)
{
    switch(event)
    {
    case EV_DUMP_FRAME:
        CamConfig_setCamReg(0x11,0x01);  /* reduce the frame rate for dumping*/
        CamConfig_sendFifoCmds();
        Utility_delay(1000);		/* allow the new frame rate to settle */
        lineCount = 0;
        currentState = ST_FrameMgr_DumpingFrame;
        CamIntAsm_waitForNewDumpFrame(currentLineBuffer,previousLineBuffer);
        break;

    case EV_ENABLE_TRACKING:
        currentState = ST_FrameMgr_TrackingFrame;
        FrameMgr_acquireFrame();
        break;

    case EV_ACQUIRE_FRAME_COMPLETE:
        FrameMgr_processFrame();
        break;

    case EV_PROCESS_FRAME_COMPLETE:
        FrameMgr_acquireFrame();
        break;

    case EV_SERIAL_DATA_RECEIVED:
        if (currentState != ST_FrameMgr_idle)
        {
            /* we need to go back to processing line data, since
                serial data reception interrupted us....just trash the
                frame and act like the frame has been processed, which
                will kick off the system to wait for the next line */
            PUBLISH_EVENT(EV_PROCESS_FRAME_COMPLETE);
        }
        break;

    case EV_DISABLE_TRACKING:
        /* tracking needs to be turned off */
        currentState = ST_FrameMgr_idle;
        break;
    }
}

/***********************************************************
    Function Name: FrameMgr_acquireFrame
    Function Description: This function is responsible for
    beginning of the acquisition of a new frame of data
    from the camera interface. The acquisition of this line
    depends on the current state of the FrameMgr.
    Inputs:  none
    Outputs: none
    ***********************************************************/
void FrameMgr_acquireFrame(void)
{
    if (currentState == ST_FrameMgr_TrackingFrame)
    {
        trackedLineCount = 0;
        numPrevTrackedObjects = numCurrTrackedObjects;
        numCurrTrackedObjects = 0;

        /* clear out the tracking table, and wait for the new frame
            to start */
        memset(trackedObjectTable,0x00,sizeof(trackedObjectTable));
        CamIntAsm_waitForNewTrackingFrame(currentLineBuffer,colorMap);
    }
}

/***********************************************************
    Function Name: FrameMgr_acquireLine
    Function Description: This function is responsible for
    acquiring a line of data from the camera interface.
    The acquisition of this line depends on the current
    state of the FrameMgr.
    Inputs:  none
    Outputs: none
    ***********************************************************/
void FrameMgr_acquireLine(void)
{
    unsigned char tmpLineCount;

    /* clearing out the buffers takes too long...we should
        just overwrite the data here without a problem when
        we start acquiring...at no point do we check for
        a 0x00 value in the current or previous lineBuffers,
        so it was a bit excessive :-)  */

    /* check which state we are in and proceed as needed */
    if (currentState == ST_FrameMgr_DumpingFrame)
    {
        tmpLineCount = lineCount*2;
        /* wait for another VSYNC so we know which frame to use
            to start looking for a line to receive */
        WAIT_FOR_VSYNC_HIGH();
        WAIT_FOR_VSYNC_LOW();

        /* look at lineCount to determine how many HREFs we should
            wait before we start sampling */
        while(tmpLineCount != 0)
        {
            WAIT_FOR_HREF_HIGH();
            tmpLineCount--;
            WAIT_FOR_HREF_LOW();
        }

        /*  we should now be ready to sample our line...*/
        CamIntAsm_acquireDumpLine(currentLineBuffer,previousLineBuffer);
    }
    else if (currentState == ST_FrameMgr_TrackingFrame)
    {
        WAIT_FOR_HREF_LOW();
        CamIntAsm_acquireTrackingLine(currentLineBuffer,colorMap);
    }
}

/***********************************************************
    Function Name: FrameMgr_processLine
    Function Description: This function is responsible for
    parsing the received image line and performing either
    connected region mapping (if in the Tracking state) or
    sending out the raw sampled data (if in the Dumping
    state).
    Inputs:  none
    Outputs: none
    ***********************************************************/
void FrameMgr_processLine(void)
{
    unsigned char i;
    volatile unsigned char dataToSend;

#ifdef DEBUG_TRACKED_LINE
    unsigned char *pSendData;
    unsigned char asciiBuffer[5];
    unsigned char pixelCount = 0;
#endif

    if (currentState == ST_FrameMgr_DumpingFrame)
    {
        /* we want to sit in a tight loop and send the acquired data
            sitting in current and previous line buffers out the serial
            port...it is sent out the serial port immediately instead
            of going into the UIMgr tx fifo because we can't do anything
            until its sent out anyway...may as well just get it out now	*/

        /* currentLineBuffer is getting "g" previousLineBuffer is getting "b-r" */
        UartInt_txByte(0x0B);			/* send the header byte */
        UartInt_txByte(lineCount);              /* send the line count */
        for (i=0; i<NUM_PIXELS_IN_A_DUMP_LINE; i+=2)
        {
            /* when a dump line is sampled, the upper byte can potentially
                have garbage in it...we don't have time to mask it off as we're
                sampling, so it is done here before we send it out...we also
                combine the samples together so we really are sending up a
                sample for line N as well as line N+1 */
            dataToSend = currentLineBuffer[i];
            dataToSend &= 0x0F;
            dataToSend <<= 4;
            dataToSend |= (previousLineBuffer[i] & 0x0F);

            /* dataToSend should be packed now */
            UartInt_txByte(dataToSend);

            /* flip the colors around since we are doing all G on Y and BR on UV */
            dataToSend = previousLineBuffer[i+1];
            dataToSend &= 0x0F;
            dataToSend <<= 4;
            dataToSend |= (currentLineBuffer[i+1] & 0x0F);

            /* dataToSend should be packed now */
            UartInt_txByte(dataToSend);
        }
        UartInt_txByte(0x0F);  /* send line end */
                               /* once all the data is sent, increment out line count by 2 since
                                   we really get 2 lines worth of pixels on each pass */
                               /* Update...increment only by 1, but only send 72 double-lines */
        lineCount++;

        /* check to see if we have retrieved all of the needed lines */
        if (lineCount >= 72)  /* half 144, since we send two lines at a time */
        {
            /* we're done, so send the dump complete?...nope, just change
                states and we should be fine */
            lineCount = 0;
            currentState = ST_FrameMgr_idle;

            /* disable the PCLK counting overflow interrupt */
            TIMSK &= DISABLE_PCLK_TIMER1_OVERFLOW_BITMASK;

            CamConfig_setCamReg(0x11,0x00);  /* reset the frame rate to normal*/
            CamConfig_sendFifoCmds();
        }
        else
        {
            /* we have more lines to acquire in this frame, so keep on truckin...*/
            PUBLISH_FAST_EVENT(FEV_PROCESS_LINE_COMPLETE);
        }
    }
    else if (currentState == ST_FrameMgr_TrackingFrame)
    {
#ifdef DEBUG_TRACKED_LINE
        /* send the received line over serial...this should only send
            until a pixelCount == 176 */
        pSendData = currentLineBuffer;
        itoa(trackedLineCount,asciiBuffer,10);
        UIMgr_txBuffer(asciiBuffer,3);
        UIMgr_txBuffer(" ",1);
        while(pixelCount < ACTUAL_NUM_PIXELS_IN_A_LINE)
        {
            memset(asciiBuffer,0x00,5);
            itoa(*pSendData++,asciiBuffer,10);	/* color is first byte */
            UIMgr_txBuffer(asciiBuffer,3);      /* 3 ascii bytes for data */
            UIMgr_txBuffer(" ",1);

            pixelCount += *pSendData;	/* run-length is second byte */
            memset(asciiBuffer,0x00,5);
            itoa(*pSendData++,asciiBuffer,10);
            UIMgr_txBuffer(asciiBuffer,3);
            UIMgr_txBuffer(" ",1);
        }
        UIMgr_txBuffer("\n\r",2);

        trackedLineCount++;
        if (trackedLineCount == 144)
        {
            UIMgr_txBuffer("  FC  \n\r",8);
            trackedLineCount = 0;
            PUBLISH_EVENT(EV_PROCESS_FRAME_COMPLETE);
        }
        else
        {
            PUBLISH_EVENT(EV_PROCESS_LINE_COMPLETE);
        }
#else
        /* determine if any of the RLE blocks overlap */
        FrameMgr_findConnectedness();
        trackedLineCount++;
        if (trackedLineCount == ACTUAL_NUM_LINES_IN_A_FRAME)
        {
            /* an entire frame of tracking data has been acquired, so
                publish an event letting the system know this fact */
            PUBLISH_EVENT(EV_ACQUIRE_FRAME_COMPLETE);
            /* disable the PCLK counting overflow interrupt */
            TIMSK &= DISABLE_PCLK_TIMER1_OVERFLOW_BITMASK;
            trackedLineCount = 0;
        }
        else
        {
            PUBLISH_FAST_EVENT(FEV_PROCESS_LINE_COMPLETE);
        }
#endif
    }
    else
    {
        /* ...and here? */
    }
}

/***********************************************************
    Function Name: FrameMgr_processFrame
    Function Description: This function is responsible for
    parsing the completed frame and performing all actions
    needed at this level.
    Inputs:  none
    Outputs: none
    ***********************************************************/
void FrameMgr_processFrame(void)
{
    unsigned char i,k,color;
#if DEBUG_FRAME_DATA
    unsigned char asciiBuffer[5];
    unsigned char j;
#endif
    unsigned char *pTableData = (unsigned char *)pCurrentTrackedObjectTable;
    unsigned char tmpUpperLeftX,tmpUpperLeftY,tmpLowerRightX,tmpLowerRightY;

#if DEBUG_FRAME_DATA
    /* we want to send all of the currently tracked table out
        the serial port for debugging */
    for (i=0; i<numCurrTrackedObjects; i++)
    {
        UIMgr_txBuffer("----------\r\n",12);
        for (j=0; j<SIZE_OF_TRACKED_OBJECT; j++)
        {
            memset(asciiBuffer,0x00,5);
            itoa(*pTableData++,asciiBuffer,10);
            UIMgr_txBuffer(asciiBuffer,3); /* 3 ascii bytes for data
                                               + 1 space */
            UIMgr_txBuffer("\r\n",2);
        }
        /* skip the fill byte in each tracked object structure */
        pTableData++;

    }

    /* finally, send a new line */
    UIMgr_txBuffer("\r\n",2);

    memset(asciiBuffer,0x00,5);
    itoa(numCurrTrackedObjects,asciiBuffer,10);
    UIMgr_txBuffer(asciiBuffer,3);
    UIMgr_txBuffer(" PFC\r\n",5);

#else
    /* we only send tracking packets if there are tracked objects */

    if (numCurrTrackedObjects > 0)
    {

        UIMgr_writeTxFifo(0x0A);					/* header byte for a tracking packet */
        UIMgr_writeTxFifo(numCurrTrackedObjects);                       /* num of objects tracked */
        for (i=0; i<numCurrTrackedObjects; i++)
        {
            /* convert the color from bit position to value...remember, each
                bit in the "color" byte corresponds to a color */
            k=0;
            color = *pTableData++;
            if (color == 128) k=0;
            else if (color == 64) k=1;
            else if (color == 32) k=2;
            else if (color == 16) k=3;
            else if (color == 8)  k=4;
            else if (color == 4)  k=5;
            else if (color == 2)  k=6;
            else if (color == 1)  k=7;

            UIMgr_writeTxFifo(k);				  	/* send the color */
            pTableData += 2;                                            /* skip the last line data */

            tmpUpperLeftX = *pTableData++;			/* get the upper left X */
            tmpUpperLeftY = *pTableData++;                      /* get the upper left Y */
            tmpLowerRightX = *pTableData++;                     /* get the lower right X */
            tmpLowerRightY = *pTableData++;                     /* get the lower right Y */

            /* skip the fill byte */
            pTableData++;

            /* calculate the center point and send it */
            UIMgr_writeTxFifo( (tmpLowerRightX+tmpUpperLeftX)>>1); 	/* send the X center point */
            UIMgr_writeTxFifo( (tmpLowerRightY+tmpUpperLeftY)>>1);      /* send the Y center point */
            UIMgr_writeTxFifo(tmpUpperLeftX);
            UIMgr_writeTxFifo(tmpUpperLeftY);
            UIMgr_writeTxFifo(tmpLowerRightX);
            UIMgr_writeTxFifo(tmpLowerRightY);
        }

        /* all done...send the end of tracking packets char */
        UIMgr_writeTxFifo(0xFF);
    }
#endif
    /* schedule the next action to acquire a new frame */
    PUBLISH_EVENT(EV_PROCESS_FRAME_COMPLETE);
}

/***********************************************************
    Function Name: FrameMgr_findConnectedness
    Function Description: This function is responsible for
    finding the connectedness between two particular run-
    length encoded lines of pixel data.  It updates the
    trackingTable as needed.
    Inputs:  none
    Outputs: none
    ***********************************************************/
static void FrameMgr_findConnectedness(void)
{
    trackedColor_t currColor;
    unsigned char *pCurrLineColorInfo = currentLineBuffer;
    unsigned char *pTrackedObjectData;
    register unsigned char currPixelRunStart=0;
    register unsigned char currPixelRunFinish=0;
    register unsigned char lastLineXStart=0;
    register unsigned char lastLineXFinish=0;
    register unsigned char runLength=1;
    unsigned char i;
    bool_t colorConnected;

    do
    {
        /* grab both the current color and the number of pixels
            in the run...remember, pixels start at 1, not 0! */
        colorConnected = FALSE;
        currColor = *pCurrLineColorInfo++;
        currPixelRunStart += runLength;
        runLength = *pCurrLineColorInfo++;
        currPixelRunFinish += runLength;

        if (currColor != notTracked)
        {
            /* this run contains a color we care about, so
                either it will begin a new tracked object, or it
                is connected to a currently tracked object...
                compare it with each object in the tracking
                table */
            for (i=0; i<numCurrTrackedObjects; i++)
            {
                /* The pTrackedObject pointer is used to move around within each
                    trackedObject_t structure to access and update its fields.  This
                    is done instead of accessing each element in the structures through
                    the more traditional way (using the 'i' index to access a structure
                    in the trackingTable, followed by accessing the particular field
                    of interest) for efficiency reasons. */
                pTrackedObjectData = (unsigned char *)&pCurrentTrackedObjectTable[i];
                if (currColor == *pTrackedObjectData)
                {
                    /* found a color match...check to see if there is
                        connectedness */
                    pTrackedObjectData++;   	/* move the pointer to the lastLineX/Y data of the structure*/
                    lastLineXStart = *pTrackedObjectData;
                    lastLineXFinish = *(pTrackedObjectData+1);

                    /* Check for the 5 following types of line connectedness:
                        ---------------------
                        |                   |
                        ---------------------
                        -------------------------
                        |                       |
                        -------------------------  */
                    if ( (	(currPixelRunStart >= lastLineXStart) &&
                          (currPixelRunStart <= lastLineXFinish) )  ||

                        /*  ---------------------
                            |                   |
                            ---------------------
                            -------------------
                            |                 |
                            -------------------
                            OR
                            ------------------------------
                            |                            |
                            ------------------------------
                            ---------
                            |       |
                            ---------  */
                        (	(currPixelRunFinish >= lastLineXStart) &&
                         (currPixelRunFinish <= lastLineXFinish) ) ||


                        /*  -------------------------------
                            |                             |
                            -------------------------------
                            -------------------------------
                            |                             |
                            -------------------------------
                            OR
                            -------------
                            |           |
                            -------------
                            -------------------------------
                            |                             |
                            -------------------------------   */
                        (  (currPixelRunStart <= lastLineXStart) &&
                         (currPixelRunFinish >= lastLineXFinish) ) )
                    {
                        /* THERE IS CONNECTEDNESS...update the lastLineXStart and lastLineXFinish
                            data pointed to by pTrackedObjectData */
                        *pTrackedObjectData++ = currPixelRunStart;
                        *pTrackedObjectData++ = currPixelRunFinish;

                        /* check if the bounding box needs to be updated */
                        if (*pTrackedObjectData > currPixelRunStart)
                        {
                            /* need to update the bounding box for the upper left point to
                                enclose this new left-most point...we never have to update the
                                upper left Y point, since each scan line we process moves from
                                top to bottom */
                            *pTrackedObjectData = currPixelRunStart;
                        }
                        pTrackedObjectData += 2;  /* move the pointer up to the lowerRight data */
                        if (*pTrackedObjectData < currPixelRunFinish)
                        {
                            /* need to update the bounding box for the lower right X point to
                                enclose this new right-most point */
                            *pTrackedObjectData = currPixelRunFinish;
                        }

                        /* the lower right 'y' point always gets updated when connectedness is found */
                        *(pTrackedObjectData+1) = trackedLineCount;

                        /* set a flag indicating that that color run is part of another
                            object and thus doesn't need to be added as a new entry into the
                            tracking table */
                        colorConnected = TRUE;
                        break;
                    }
                }
            }

            if (colorConnected == FALSE)
            {
                /* a new entry needs to be made to the tracking table, since we have
                    a run-length with a color, and it isn't connected to anything...but we
                    can only do this if there is space left in the trackedObject table */
                if (numCurrTrackedObjects < MAX_TRACKED_OBJECTS)
                {
                    /* space is available...add the object */
                    pTrackedObjectData = (unsigned char *)&pCurrentTrackedObjectTable[numCurrTrackedObjects];

                    /* now that we have a pointer to the tracked object to be updated, update all
                        the fields */
                    *pTrackedObjectData++ = currColor;				  /* color */
                    *pTrackedObjectData++ = currPixelRunStart;                  /* lastLineXStart */
                    *pTrackedObjectData++ = currPixelRunFinish;                 /* lastLineXFinish */
                    *pTrackedObjectData++ = currPixelRunStart;                  /* x_upperLeft */
                    *pTrackedObjectData++ = trackedLineCount;                   /* y_upperLeft */
                    *pTrackedObjectData++ = currPixelRunFinish;                 /* x_lowerRight */
                    *pTrackedObjectData++ = trackedLineCount;                   /* y_lowerRight */
                                                                                /* we'll calculate the center point for each tracked object when
                                                                                    the frame is done, instead of wasting the precious clock cycles doing
                                                                                    it now */

                    numCurrTrackedObjects++;
                }
            }
        }
    } while(currPixelRunFinish < ACTUAL_NUM_PIXELS_IN_A_LINE);
}

