package img.connection;

import java.io.*;
import java.net.*;
import java.nio.channels.*;

/***********************************************************************
 * This class represents a Connection to a Datagram Socket.
 */
public class DatagramSocketConnection extends AbstractConnection {

   /**
    * The Datagram Channel to receive/send data from/to.
    */
   private DatagramChannel channel;

   /**
    * The port number for this datagram channel to bind to on the local host.
    */
   private int port;

   /**
    * Create a Datagram Channel to bind to the given port on the local host.
    * @param port The port number to listen for incoming data on.
    */
   public DatagramSocketConnection(int port) {
      super();

      this.port = port;
   }

   /**
    * Returns the DatagramChannel object associated with this connection.
    * @return The DatagramChannel.
    */
   public Object getConnectionObject() {
      return channel;
   }

   /**
    * Open the Datagram Channel and bind to a port on the local host and
    * fire a connected event.
    * @throws IOException If the channel can not be opened.
    */
   public void connect() throws IOException {
      SocketAddress address = new InetSocketAddress(InetAddress.getLocalHost(),
                                                    port);
      channel = DatagramChannel.open();
//      channel.socket().setReceiveBufferSize(8000);
      channel.socket().bind(address);
      setConnected();
   }

   /**
    * Close the Datagram Channel and fire a disconnected event.
    * @throws IOException If the channel can not be closed.
    */
   public void disconnect() throws IOException {
      channel.close();
      setDisconnected();
   }

   /**
    * Returns the ByteChannel associated with the open Datagram Channel
    * @return The ByteChannel to read and write data.
    */
   public ByteChannel getChannel() {
      return channel;
   }

   /**
    * Returns an input stream for this Connection.
    * @return An input stream for reading bytes from this Connection.
    * @throws IOException
    */
   public InputStream getInputStream() throws IOException {
      return Channels.newInputStream(channel);
   }

   /**
    * Returns an output stream for this Connection.
    * @return An output stream for writing bytes to this Connection.
    * @throws IOException
    */
   public OutputStream getOutputStream() throws IOException {
      return Channels.newOutputStream(channel);
   }

}
