package img.connection;

import java.io.*;
import java.net.*;
import java.nio.*;
import java.nio.channels.*;

/***********************************************************************
 * This class represents a Connection to a Socket.
 */
public class SocketConnection extends AbstractConnection {

   /**
    * The number of milliseconds to wait for a connection to be made.
    */
   private static final int CONNECTION_TIMEOUT = 2000;

   /**
    * The Socket object through which this Connections is attached to.
    */
   private Socket socket;

   /**
    * The SocketAddress this Connection is going to connect to.
    */
   private InetSocketAddress address;

   /**
    * The amount of time (in milliseconds) to wait for a connection
    * to be established.
    */
   private int timeout;

   /**
    * Create a new SocketConnection to the given host name, port, and 2
    * second timeout.
    * @param host The name of the host to connect to.
    * @param port The port number to use.
    */
   public SocketConnection(String host, int port) {
      // use a default timeout of 2 seconds.
      this(new InetSocketAddress(host, port));
   }

   /**
    * Create a new SocketConnection to the given Socket Address, with a 2
    * second timeout.
    * @param address The Socket Address to connect to.
    */
   public SocketConnection(InetSocketAddress address) {
      this(address, CONNECTION_TIMEOUT);
   }

   /**
    * Create a new SocketConnection to the given Socket Address and timeout
    * value.
    * @param address The Socket Address to connect to.
    * @param timeout The time to wait for a connection to open.
    */
   public SocketConnection(InetSocketAddress address, int timeout) {
      super();

      System.out.println("Connect to: " + address);

      this.address = address;
      this.timeout = timeout;
   }

   /**
    * Get the Socket object through which this Connection is wrapped around.
    * @return The Socket through which this Connection is wrapped around.
    */
   public Object getConnectionObject() {
      return socket;
   }

   /**
    * Connects this Connection to the given SocketAddress.
    * @throws IOException If the connection can not be established.
    */
   public synchronized void connect() throws IOException {
      socket = new Socket(address.getAddress(), address.getPort());
      socket.setSendBufferSize(8192);
      System.out.println("Send Buffer Size: " + socket.getSendBufferSize());
      setConnected();
   }

   /**
    * Disconnects this Connection.
    * @throws IOException If the connection can not be destroyed.
    */
   public synchronized void disconnect() throws IOException {
      socket.close();
      setDisconnected();
   }

   /**
    * Returns an input stream for this Connection.
    * @return An input stream for reading bytes from this Connection.
    * @throws IOException
    */
   public InputStream getInputStream() throws IOException {
      return socket.getInputStream();
   }

   /**
    * Returns an output stream for this Connection.
    * @return An output stream for writing bytes to this Connection.
    * @throws IOException
    */
   public OutputStream getOutputStream() throws IOException {
      return socket.getOutputStream();
   }

   /**
    * Returns the unique ByteChannel object associated with this Connection.
    * @return The byte channel associated with this Connection
    */
   public ByteChannel getChannel() {
      try {
         return new SocketByteChannel(getInputStream(), getOutputStream());
      } catch(IOException ex) {
         return null;
      }
   }

   /**
    * Returns a string representation of this Connection.
    * @return A string representation of this Connection.
    */
   public String toString() {
      return socket.toString();
   }

   /***
    * This class is a wrapper/adapter around the Input and Output Streams
    * of the Socket so they can be used with the New I/O in Java 2 Version 1.4.
    */
   private class SocketByteChannel implements ByteChannel {

      /**
       * The Byte Channel to read bytes from.
       */
      private ReadableByteChannel in;

      /**
       * The Byte Channel to write bytes to.
       */
      private WritableByteChannel out;

      /**
       * Construct a new SocketByteChannel for reading from in and writing
       * to out.
       * @param in The Byte Channel to read bytes from.
       * @param out The Byte Channel to write bytes to.
       */
      public SocketByteChannel(InputStream in, OutputStream out) {
         this.in = Channels.newChannel(in);
         this.out = Channels.newChannel(out);
      }

      /**
       * Closes this channel.
       * @throws IOException if the input and output stream could not be closed
       */
      public void close() throws IOException {
         in.close();
         out.close();
      }

      /**
       * Tells whether or not this channel is open.
       * @return true if, and only if, this channel is open
       */
      public boolean isOpen() {
         return in.isOpen() && out.isOpen();
      }

      /**
       * Writes a sequence of bytes to this channel from the given buffer.
       * @param byteBuffer The buffer from which bytes are to be retrieved
       * @return The number of bytes written, possibly zero
       * @throws IOException If some other I/O error occurs
       */
      public int write(ByteBuffer byteBuffer) throws IOException {
         return out.write(byteBuffer);
      }

      /**
       * Reads a sequence of bytes from this channel into the given buffer.
       * @param byteBuffer The buffer into which bytes are to be transferred
       * @return The number of bytes read, possibly zero, or -1 if the
       * channel has reached end-of-stream
       * @throws IOException If some other I/O error occurs
       */
      public int read(ByteBuffer byteBuffer) throws IOException {
         return in.read(byteBuffer);
      }
   }

}
